import typing
import random

from . import http_client, enumeration, result

Tag = typing.Union[str, enumeration.SFWImageTags, enumeration.NSFWImageTags, enumeration.RealSFWImageTags]


class NekosLifeClient:
    """
    Main client that makes requests and returns objects for better handling.

    Parameters
    ----------
    session : asyncio.ClientSession
    """
    def __init__(self, *, session=None):
        self.http = http_client.HttpClient(session=session)

    # V2
    async def image(self, tag: typing.Union[str, enumeration.SFWImageTags, enumeration.NSFWImageTags, enumeration.RealSFWImageTags], get_bytes: bool=False):
        """
        -> Coroutine
        Returns an image of a specific tag.

        Parameters
        ----------
        tag : Union[str, anekos.SFWImageTags, anekos.NSFWImageTags]
            The tag of the image.
        get_bytes : bool (optional)
            Gets the bytes of an image.
            You can get the bytes of the image by accessing the `bytes` attribute  of the returned object.

        Returns
        -------
        anekos.result.ImageResult
        """
        if not isinstance(tag, (str, enumeration.SFWImageTags, enumeration.NSFWImageTags, enumeration.RealSFWImageTags)):
            raise TypeError(f"{Tag} expected in `tag`")

        if type(get_bytes) is not bool:
            raise TypeError("bool expected in `get_bytes`")

        str_tag = tag if type(tag) is str else tag.value

        data_response = await self.http.endpoint("img/" + str_tag)

        data_response["tag"] = tag

        if get_bytes:
            image_url = data_response["url"]
            image_bytes = await self.http.get_image_bytes(image_url)
            data_response["bytes"] = image_bytes

        return result.ImageResult(data_response)

    async def random_fact_text(self):
        """
        -> Coroutine
        Returns a random fact.

        Returns
        -------
        anekos.result.TextResult
        """
        data_response = await self.http.endpoint("fact")
        return result.TextResult(data_response, target="fact")

    async def random_cat_text(self):
        """
        -> Coroutine
        Returns a random cat-like text.

        Returns
        -------
        anekos.result.TextResult
        """
        data_response = await self.http.endpoint("cat")
        return result.TextResult(data_response, target="cat")

    async def random_why(self):
        """
        -> Coroutine
        Returns a random questionnaire.

        Returns
        -------
        anekos.result.TextResult
        """
        data_response = await self.http.endpoint("why")
        return result.TextResult(data_response, target="why")

    async def spoiler(self, text: str):
        """
        -> Coroutine
        Creates an individual spoiler per letter for Discord.

        Paramaters
        ----------
        text : str
            It must be between 1 and 2000 characters.

        Returns
        -------
        anekos.result.TextResult
        """
        if type(text) is not str:
            raise TypeError("str expected in `text`")

        if not (0 < len(text) < 2000):
            raise TypeError("'text' must be between 1 and 2000 characters")

        data_response = await self.http.endpoint("spoiler", text=text)
        return result.TextResult(data_response, target="owo")

    async def random_name(self):
        """
        -> Coroutine
        Returns a random name.

        Returns
        -------
        anekos.result.TextResult
        """
        data_response = await self.http.endpoint("name")
        return result.TextResult(data_response, target="name")

    async def owoify(self, text: str):
        """
        -> Coroutine
        OwOify the `text`.

        Parameters
        text : str

        Returns
        -------
        anekos.result.TextResult
        """
        if type(text) is not str:
            raise TypeError("str expected in `text` parameter")

        if not (0 < len(text) < 200):
            raise IndexError("'text' must be between 1 or 200 characters")

        data_response = await self.http.endpoint("owoify", text=text)
        return result.TextResult(data_response, target="owo")

    async def random_8ball(self, question, *, get_image_bytes: bool=False):
        """
        -> Coroutine
        Returns a random answer for the specified question.

        Parameters
        ----------
        question : str
            The your question.
        get_image_bytes : bool
            Gets the bytes of the image.

        Returns
        -------
        anekos.result.EightBallResult
        """
        if type(question) is not str:
            raise TypeError("str expected in `question` parameter")

        if type(get_image_bytes) is not bool:
            raise TypeError("bool expected in `get_image_bytes` parameter")

        data_response = await self.http.endpoint("8ball")

        if get_image_bytes:
            image_url = data_response["url"]
            image_bytes = await self.http.get_image_bytes(image_url)
            data_response["image_bytes"] = image_bytes

        return result.EightBallResult(data_response)

    async def endpoints(self):
        """Alias for anekos.HTTPClient.endpoints."""
        return await self.http.endpoints()



"""
Async Nekos.Life API Wrapper

A basic wrapper for Nekos.Life API.
"""

__title__ = "anekos"
__author__ = "NiumXp"
__license__ = "MIT"
__version__ = "1.0.0"

from .client import NekosLifeClient
from .enumeration import NSFWImageTags, SFWImageTags, RealSFWImageTags
from . import result, errors
from .http_client import Endpoint



from enum import Enum
from typing import List


class AutoName(Enum):
    def _generate_next_value_(name, *args):
        # https://docs.python.org/3/library/enum.html#using-automatic-values
        return name.lower()

    @classmethod
    def to_list(cls) -> List[str]:
        """Get all values, put them into a list and return.

        Returns
        -------
        List[str]
        """
        return [attr.value for attr in cls]


class NSFWImageTags(AutoName):
    LEWD = "lewd"
    GASM = "gasm"
    SPANK = "spank"


class SFWImageTags(AutoName):
    WALLPAPER = "wallpaper"
    NGIF = "ngif"
    TICKLE = "tickle"
    FEED = "feed"
    GECG = "gecg"
    SLAP = "slap"
    AVATAR = "avatar"
    WAIFU = "waifu"
    PAT = "pat"
    KISS = "kiss"
    NEKO = "neko"
    CUDDLE = "cuddle"
    FOX_GIRL = "fox_girl"
    HUG = "hug"
    SMUG = "smug"


class RealSFWImageTags(AutoName):
    LIZARD = "lizard"
    EIGHT_BALL = "8ball"
    GOOSE = "goose"
    WOOF = "woof"



from aiohttp import ClientSession

from urllib.parse import urljoin, quote
from collections import namedtuple

from . import errors


base = "https://nekos.life/api/v2/"
Endpoint = namedtuple("Endpoint", "methods url is_deprecated")


def _make_url(endpoint, parameters):
    url = urljoin(base, endpoint)
    if parameters:
        url += '?' + '&'.join(f"{k}={quote(v)}" for (k,v) in parameters.items())
    return url


class HttpClient:
    """Client for making API requests.

    Parameters
    ----------
    session : asyncio.ClientSession
    """
    def __init__(self, *, session=None):
        if session:
            self._session = session
        else:
            self._session = ClientSession()

        self._endpoints = None

    async def get(self, endpoint: str, **parameters):
        url = _make_url(endpoint, parameters)
        async with self._session.get(url) as response:
            status = response.status

            if status in [500, 404]:
                raise errors.NoResponse("endpoint not found")

            response = await response.json()
            # the response can be a list or a dict
            # a list when the request is in /api/v2/endpoints

            if type(response) is dict:
                if response.get("msg", None):
                    raise errors.NoResponse("endpoint not found")

        return url, response
 
    async def get_image_bytes(self, url: str):
        """
        -> Coroutine
        Returns the image in bytes.

        Paremeters
        ----------
        url : str
            Image URL.

        Returns
        -------
        bytes
            Image in bytes.
        """
        async with self._session.get(url) as response:
            return await response.read()
 
    async def endpoint(self, endpoint: str, **parameters):
        """
        -> Coroutine
        Makes a GET request in the API's `endpoint`.

        Parameters
        ----------
        endpoint : str
            The endpoint where the GET request will be made.

        Returns
        -------
        dict
            The result of the request made.
        """
        url, response = await self.get(endpoint, **parameters)
        return {"endpoint": url, **response}

    async def endpoints(self):
        """
        -> Coroutine
        Returns all endpoints of the API.

        Returns
        -------
        List[anekos.Endpoint]
        """
        if self._endpoints:
            return self._endpoints

        _, endpoints_ = await self.get("endpoints")
        for _ in range(len(endpoints_)):
            endpoint = endpoints_.pop(-1)
            methods, right = endpoint.split('     ')

            if right.endswith("-DEPRECATED"):
                is_deprecated = True
                url, _ = right.split(' ', 1)
            else:
                is_deprecated = False
                url = right

            url = "https://nekos.life" + url

            endpoint = Endpoint(methods.split(','), url, is_deprecated)
            endpoints_.insert(0, endpoint)

        self._endpoints = endpoints_
        return endpoints_



from os.path import splitext, basename, join

#from .enumeration import SFWImageTags, NSFWImageTags

try:
    from aiofile import AIOFile
except ModuleNotFoundError:
    has_aiofile = False
else:
    has_aiofile = True


class Result:
    """
    Base structure for API returns.

    Attributes
    ----------
    endpoint : str
        The endpoint in which the result was returned.
    """
    def __init__(self, data: dict):
        self.endpoint = data.pop("endpoint")


class TextResult(Result):
    """
    Structure used for endpoints that return texts.

    Attributes
    ----------
    endpoint : str
        The endpoint in which the result was returned.
    text : str
        The returned text.
    """
    def __init__(self, data: dict, target: str="text"):
        super().__init__(data)

        self.text = data.pop(target)


class ImageResult(Result):
    """
    Structure used for endpoints that return images.

    Attributes
    ----------
    endpoint : str
        The endpoint in which the result was returned.
    url : str
        Image URL.
    full_name : str
        Full image name.
    name : str
        Image name (without the extension).
    extension : str
        Image extension (PNG, JPG, ...).
    tag : Union[str, anekos.NSFWImageTags, anekos.SFWImageTags]
    bytes : bytes
        Image in bytes, it can be `None`.
    """
    def __init__(self, data: dict):
        super().__init__(data)
        
        self.url = url = data.pop("url")
        self.full_name = full_name = basename(url)
        self.name, self.extension = splitext(full_name)
        self.tag = data.pop("tag")
        self.bytes = data.pop("bytes", None)

    def sync_save(self, path: str):
        """Saves the image in `path`.

        Parameters
        ----------
        path : str
            The path there image will be saved.

        Notes
        -----
        THIS FUNCTION CAN "LOCK" YOUR CODE.
        TRY USE `save` METHOD INSTEAD.
        """
        path = join(path, self.full_name)
        with open(path, "wb") as file:
            try:
                file.write(self.bytes)
            except TypeError:
                raise RuntimeError("you need to get the image bytes to use this method")

    def sync_download(self, path: str):
        """Alias for `sync_save` method."""
        return self.sync_save()

    async def save(self, path: str):
        """Saves the image in the `path`.

        Parameters
        ----------
        path : str
            The path where the image will be saved.

        Notes
        -----
        THE `aiofile` LIBRARY IS NEEDED IN ORDER TO USE THIS METHOD.
        USE `sync_save` IF YOU DO NOT WANT TO INSTALL THE `aiofile` LIBRARY.
        """
        if not has_aiofile:
            raise RuntimeError("aiofile library is needed in order to use this method.")

        path = join(path, self.full_name)
        async with AIOFile(path, "wb") as file:
            try:
                await file.write(self.bytes)
            except ValueError:
                raise RuntimeError("you need to get the image bytes to use this method")

    async def download(self, path: str):
        """Alias for `save` method."""
        return await self.save(path)


class EightBallResult(Result):
    """
    Structure used for the endpoint `/8ball`

    Attributes
    ----------
    endpoint : str
        The endpoint in which the result was returned.
    text : str
        The returned text.
    image_url : str
        Returned image URL.
    image_bytes : bytes
        Image in bytes, it can be `None`.
    """
    def __init__(self, data: dict):
        super().__init__(data)

        self.text = data.pop("response")
        self.image_url = data.pop("url")
        self.image_bytes = data.pop("image_bytes", None)




from setuptools import setup

setup(name="anekos.py",
      author="regulad",
      url="https://github.com/regulad/anekos.py",
      version="2.0.0",
      packages=["anekos"],
      install_requires=["aiohttp>=3.6.2"],
      license="MIT",
      description="An unofficial async wrapper for nekos.life API!",
      python_requires=">=3.6"
)
